(*
Authors:

    Bruno Julia-Diaz, Email : bjulia@gmail.com,
    Department of Physics & Astronomy,
    University of Pittsburgh, Pittsburgh, PA 15260
    
    Frank Tabakin, Email : tabakin@pitt.edu,
    Department of Physics & Astronomy,
    University of Pittsburgh, Pittsburgh, PA 15260

version 0.0, 10 Aug 2005
*)

    
 
BeginPackage["QDENSITY`Qdensity`", {"LinearAlgebra`MatrixManipulation`"}]
(*Begin["`Private`"];*)
rcsRevision = StringReplace["$Revision: 1.00 $", "$" -> ""];
rcsDate     = StringReplace["$Date: 8-3-2005 1:00:00  $", "$" -> ""];
(*End[];*)

Print["Type \"?intro\", \"?QDENSITY`Qdensity`*\" or \"?about\ for help."];

$PrePrint = MatrixForm; 
Off[General::"spell1"]

(* Begins USAGE description of Commands *)
about::usage = 
" Qdensity is a Mathematica package to simulate a quantum computer 
\n on a classical one. This simulation closely follows the procedures 
\n that would occur in an actual quantum computer, wherein quantum 
\n interference and superposition play a key role.  
\n Although the main emphasis is on using a density matrix approach, 
\n QDENSITY also handles state vectors. The package comes with a manual 
\n and several notebooks that should allow the user to build any quantum 
\n circuit out of conventional quantum gates.  

\n\n The notebooks develop the most relevant quantum algorithms: 
\n teleportation, searching algorithms and factoring algorithms in 
\n a self contained way. Some basic cluster model examples are also provided.

\n\n We thank Joseph Burdis for his help with QDENSITY. We also acknowledge 
\n the participation of Matt McHugh and Mahesh Bandi at an early stage of 
\n this project. Reza Yoosoofmiya and Sushmita Biswas provided some helpful 
\n suggestions on the final versions. Insightful comments by Profs. E. Gerjuoy (PITT) 
\n and Robert B. Griffiths (CMU) are very much appreciated. 
\n This project was supported,in part, by the N.S.F ";

intro::usage =
"The commonly used quantum gates and structures used in quantum
\n circuits are implemented in the package. Several examples including 
\n teleportation circuits, Grover search algorithms and Shor's 
\n factoring algorightms are included. Several additional applications are are illustrated."
 
(*  ================================================================= *)
(*  START USAGE DOCUMENTATION  *)

Adj::usage=
"Adj[A]
\n\t gives the adjoint of matrix A.";
 
CircleTimes::usage=
"\[CircleTimes] 
\n used as A \[CircleTimes] B is the tensor product of A and B.
\n When nested, care should be taken with the use of parentheses
\n e.g. use (A \[CircleTimes] B) \[CircleTimes] C instead of simply 
\n A \[CircleTimes] B \[CircleTimes] C.";

Fidelity::usage=
"Fidelity[Op1,Op2]
\n\t returns the concurrence of operators Op1 and Op2
\n defined as =MatrixPower[ MatrixPower[Op1,1/2].Op2.MatrixPower[Op1,1/2],1/2].";

ControlledX::usage=
"ControlledX[L,q1,q2]
\n\t Returns the ControlledX operator with q1 qubit as control, 
\n q2 qubit as target within L qubits--same as CNOT ";

ControlledY::usage=
"ControlledY[L,q1,q2]
\n\t Returns the ControlledY operator with q1 qubit as control,
\n q2 qubit as target within L qubits";

ControlledZ::usage=
"ControlledZ[L,q1,q2]
\n\t Returns the ControlledY operator with q1 qubit as control,
\n q2 qubit as target within L qubits";

CNOT::usage=
"CNOT[L,q1,q2]
\n\t Returns the CNOT operator with q1 qubit as control, 
\n q2 qubit as target within L qubits--same as ControlledX. ";

Comm::usage =
"Comm[A,B]
\n\t gives the commutator of A and B defined as AB-BA.";

CPHASE::usage=
"CPHASE[L,q1,q2]
\n\t Returns the CPHASE operator of q1 and q2 within L qubits. 
\n Its action on twoqubits is |ab\[RightAngleBracket]-->(-)^(ab)|ab\[RightAngleBracket]
\n --same as ControlledZ.";

CPHASEV::usage=
"CPHASEV[q1,q2,A],
\n\t returns the CPHASE operated on the vector A, which should be a list of Lenght 2^n.";


Entropy::usage="
Entropy[\[Rho]],
\n\t gives the entropy of the system described by the 
\n density matrix \[Rho]. Defined = -Tr[ \[Rho].Log[\[Rho],2]].";

\[ScriptCapitalH]::usage=
"The single qubit Hadamard--See Hadamard";

had::usage=
"had[L, q]  where L=n_q is the total number of qubits
 and q designates the single qubit on which a Hadamard acts--see Hadamard.";

Had::usage=
"see Hadamard.";

HALL::usage=
"HALL[L]
\n\t returns a Hadamard applied to each of the L qubits
\n of a L qubit system-see Hadamard.";

Hadamard::usage=
"Hadamard, had, Had, HALL:
\n\t \[ScriptCapitalH]
\n Hadamard  acting on one qubit.
\n\t had[L,q] is a Hadamard acting on one qubit q out of L 
\n qubits (use as had[1,1] or had[2,2]).
\n\t Had[L,Q]   acts on  all members of set 
\n Q={q1,q2,...} of the L qubits,  where qi>0 indicated that
\n the ith qubit is acted on by a Hadamard, whereas qj=0 
\n designates that the jth qubit is acted on by a unit 2x2 operator.
\n\t For example, Had[3,{1,0,1}] has a Hadamard acting on 
\n qubits 1 and 3 and a unit 2x2 acts on qubit 2.
\n To get a Hadamard acting on all qubits, include all 
\n qubits in Q, e.g., use Q={1,1,1,....}.
\n\t HALL[L]=Had[L,{1,1,1....}] is also provided where 
\n the array of 1's has length L of all the qubits";

\[DoubleStruckCapitalI]::usage="
\[DoubleStruckCapitalI]
\n returns the identity matrix for one qubit, 2x2.";

Ket::usage=
"Ket[0], Ket[1],
\n\t one qubit |0\[RightAngleBracket] and |1\[RightAngleBracket] state, respectively.";

Ket1::usage=
"Ket1[\[Theta],\[Phi]],
\n\t returns a general one qubit ket state of the form
\n =Cos[\[Theta]/2]*Exp[-\[ImaginaryI]*\[Phi]/2]*Ket[0]+Sin[\[Theta]/2]*Ket[1]*Exp[-\[ImaginaryI]*\[Phi]/2]  ";

KetX::usage=
"KetX[0], KetX[1], 
\n\t one qubit |+\[RightAngleBracket], |-\[RightAngleBracket], states, respectively. 
\n Eigenvectors of Sigma1.";

KetY::usage=
"KetY[0], KetY[1], 
\n Eigenvectors of Sigma2.";

KetV::usage=
"KetV[a_] is a multi-qubit state vector ket,
\n\t variable a must be list of Length[a]=2^n. Example of 
use, KetV[{0,1,1}]=(Ket[0]\[CircleTimes]Ket[1])\[CircleTimes]Ket[1].";

Bra::usage=
"Bra[0], Bra[1],
\n\t one qubit \[LeftAngleBracket]0|, \[LeftAngleBracket]1|, state, respectively.";

Bra1::usage=
"Bra1[\[Theta],\[Phi]],
\n\t returns a general one qubit bra state.";

BraX::usage=
"BraX[0], BraX[1],
\n\t one qubit \[LeftAngleBracket]+|, \[LeftAngleBracket]-|, respectively.";

BraV::usage=
"BraV[a_] is a multi-qubit state vector bra,
\n\t variable a must be list.
\n\t Example of use, 
BraV[{a,b,c}]=(Bra[a]\[CircleTimes]Bra[b])\[CircleTimes]Bra[c],
with a,b and c {0,1}";

MeasureKet::usage=
"MeasureKet[V,M,MV],
\n\t Performs the measurement on qubits in list M={m1,..,mn},
\n with possible measured computational basis values input in list 
\n MV={0,1,...0,0,1}, on the original ket V. The output is normalized 
\n to 1.";

CompMeasureKet::usage=
"Precompiled version of MeasureKet.";

MultiProjCBasis::usage=
"MultiProjCbasis[{a1,a2,..,an}]
\n\t ai={0,1} builds the multi-qubit projector operator in the computational basis: \[ScriptCapitalP]a1 \[CircleTimes] \[ScriptCapitalP]an";

MultiQProj::usage=
"MultiQProj[{a1,a2,..,an},{b1,b2,...,bn}]
\n\t  ai={0,1,2,3} and bi={0,1} builds the following multi-qubit projector 
\n operator: 
\n    ai correspond to: (0) no measurement,(1) X, (2) Y and (3) Z.
\n    bi correspond to the measured values in the same order. 
\n The length of both lists should be the same. The values b 
\n for the unmeasured qubits are not relevant.
\n
\n Example:
\n (((\[ScriptCapitalP]0 \[CircleTimes] \[DoubleStruckCapitalI])\[CircleTimes] PX[b])
\n
\n corresponds to MultiQProj[{3,0,1},{0,0,b}]
\n where the second 0 in the measured values matrix is 
\n not taken into account.";


nq::usage = 
"Returns the total number of qubits.";

nb::usage = 
"Sets the basis to 2, i.e 0 and 1.";

Oracle::usage=
"Oracle[nqx,nqy,itype]
\n\nis an Oracle for nqx Qubits in x and nqy Qubits in y. 
\n Oracle | x, y\[RightAngleBracket] = | x, y+f(x)\[RightAngleBracket].
\n
\n Oracle[nqx,nqy,itype] provides an Oracle type
\n itype=-1 ( a random function f(x)) and itype=-2 
\n ( a f(x) with value 1 only at \
\n a random marked value of x).
\n The later is useful for Grover's search. 
\n See also Definef";

\[ScriptCapitalP]::usage=
"Subscripted with 0 or 1 returns the one qubit projection operator
\n |0\[RightAngleBracket]\[LeftAngleBracket]0| or |1\[RightAngleBracket]\[LeftAngleBracket]1| respectively.
\n Commonly used together with \[CircleTimes] to 
build density matrices.";

CompPTr::usage="
Compiled version of PTr, see PartialTrace.";

PTr::usage=
"See PartialTrace.";

PartialTrace::usage=
"PTrS[q,L,V] and PTr[traceout,\[CapitalOmega]] 
\n
\n PTr[traceout,\[CapitalOmega]] returns the partial trace 
\n of general operator \[CapitalOmega] over those qubits 
\n specified by the traceout array 'traceout'. 
\n
\n PTrS[q, L,V] returns the partial trace of spin basis 
\n operator of length L over single qubit q within L qubits, 
\n with the array V specifying the spin basis operator.
\n This method reduces the operator down to the requisite smaller arrays.
\n
\n  see also PTrGen";

PartialTraceGen::usage=
"PTrGen[traceout,\[CapitalOmega]] 
\n
\n PTrGen[traceout,\[CapitalOmega]] returns the partial 
\n trace of a general operator \[CapitalOmega] over those
\n qubits specified by the traceout array 'traceout'.  
This method builds up the requisite arrays and can be faster. 
\n
\n see also PTr";

PX::usage= 
"PX[a], 
\n\t Projection operators onto the eigenstates 
of SigmaX. PX[0] and PX[1] are equivalent to |+\[RightAngleBracket]\[LeftAngleBracket]+| and 
|-\[RightAngleBracket]\[LeftAngleBracket]-| respectively.";

PY::usage= 
"PY[a], 
\n\t Projection operators onto the eigenstates 
of SigmaY.";
 
ProjG::usage=
"ProjG[a,vec], 
\n\t General Projection operator into any direction defined by the three vector vec. 
\n\t To ensure that the result 
is a projection operator the vector vec should have |vec|=1.";

Purity::usage=
"Purity[\[Rho]]
\n\t returns the purity of the state represented by the 
\n density matrix \[Rho]. It is simply Tr[\[Rho].\[Rho]].
\n In general the Purity is a real number between 0 and 1, 
\n being equal to 1 only if the state is a pure state.";

RandomKet1::usage=
"RandomKet1, 
\n\t returns a random one qubit ket.";

RandomBra1::usage=
"RandomBra1, 
\n\t returns a random one qubit bra.";

RandomQubit1::usage=
"RandomQubit1, 
\n\t returns a random density matrix for one 
Qubit.";

RandomQubit2::usage=
"RandomQubit2, 
\n\t returns a random density matrix for a two qubit
state.";

RandomQubitN::usage=
"RandomQubitN[L], 
\n\t returns a random density matrix for a L qubit
state.";

Rotqbit::usage=
"Rotqbit[v,t],
\n\t builds the matrix corresponding to a rotation around 
\n axis v of an angle t: MatrixExp[-\[ImaginaryI] t/2 \[Sigma].v]
"; 

RotX::usage=
"RotX[t], 
\n\t Rotation around the X axis by an angle t.";

RotY::usage=
"RotY[t], 
\n\t Rotation around the Y axis by an angle t.";

RotZ::usage=
"RotZ[t], 
\n\t Rotation around the Z axis by an angle t.";

\[Sigma]::usage =
"Should be used with subscripts 0,1,2,3. Returns the 
corresponding Pauli Matrix.";

Sigma0::usage=
"Sigma0
\n returns the 2x2 identity matrix.";

Sigma1::usage=
"Sigma1
\n returns the Pauli sigma x matrix.";

Sigma2::usage=
"Sigma2
\n returns the Pauli sigma y matrix.";

Sigma3::usage=
"Sigma3
\n returns the Pauli sigma z matrix.";

s::usage=
"s[i] with i={0,1,2,3}
\n Returns the Pauli Sigma matrices.";

SP::usage=
"SP[Length[V],V]
\n\t SP returns the tensor or spin product(SP) of 
\n Length[V] qubits with the components
\n specified by the array V.  The array V={0, 1,2,q,...}has 
\n a length equal to the number of qubits
\n and the entries in the array V specify the
\n components of the qth Pauli operator.  
\n\t For example:  SP[2,{3,2}] returns the same 4X4 matrix as
\n Sigma3\[CircleTimes]Sigma2.  The advantage of SP is that it
\n works for large dimensional matrices.";

Swap::usage=
"Swap[L,q1,q2]
\n\t swaps the content of qubits 1 and 2,  within L qubits.";

TensorProduct::usage =
"A\[CircleTimes]B,
\n\t gives the tensor product of A and B. 
\n If concatenated special care should be paid to the use 
\n of parentheses, e.g., A\[CircleTimes]B\[CircleTimes]C 
\n would NOT work properly, use (A\[CircleTimes]B)\[CircleTimes]C instead.";

ThreeOp::usage=
"ThreeOp[L_,q1_,q2_,q3_,op1_,op2_,op3_]
\n\t used to 
build three body operators from the Pauli matrices in a multiqubit system. It builds 
operation \[Sigma]op1 on q1, \[Sigma]op2 on q2 and \[Sigma]op3 on q3.";

Toffoli::usage=
"Toffoli[L,q1,q2, q3] 
\n\tReturns the Toffoli operator with q1 and q2 as control qubits, 
\n q3 as target qubit within L qubits ";

TwoOp::usage=
"TwoOp[L_,q1_,q2_,op1_,op2_]
\n\t is the main ingredient used to 
build two body operators from the Pauli matrices in a multiqubit system. It builds 
operation \[Sigma]op1 on q1 and \[Sigma]op2 on q2.";

xb::usage = 
"xb[i]
\n\t returns number i in binary;  
\n See Xb (note capital X) for the associated binary array from 0 to nb^nq -1.";

xd::usage = 
"xd[i]
\n\t returns decimal number i.
\n See Xd (note capital X) for the assocated decimal array from 0 to nb^nq -1.";

Xb::usage = 
"Xb (note capital X)
\n\t returns the binary 
array for the numbers 0 to nb^nq -1.";

Xd::usage = 
"Xd (note capital X)
\n\t the decimal array for 
the numbers 0 to nb^nq -1.";

ZeroT::usage=
"ZeroT[L]
\n returns a table of 0 of length L.";

(*  END USAGE DOCUMENTATION  *)
(*  ================================================================= *)

(* Preset base as binary nb=2, and nq=6 qubits *)
nb=2;
nq:=6;

(*  START DEFINITIONS *)
(*  Define qubit arrays *)

Ket[0]:={{1},{0}};
Ket[1]:={{0},{1}};
KetX[0]:=1/Sqrt[2]*{{1},{1}};
KetX[1]:=1/Sqrt[2]*{{1},{-1}};
KetY[0]:=1/Sqrt[2]*{{1},{\[ImaginaryI]}};
KetY[1]:=1/Sqrt[2]*{{1},{-\[ImaginaryI]}};
Bra[0]:=Adj[Ket[0]];
Bra[1]:=Adj[Ket[1]];
BraX[0]:=Adj[KetX[0]];
BraX[1]:=Adj[KetX[1]];
BraY[0]:=Adj[KetY[0]];
BraY[1]:=Adj[KetY[1]];


(* Begins the private environment *) 
Begin["Private`"]

Q:=Array[q, nq];

(* Multiqubit Ket and Multiqubit Bra *)

KetV[a_?ListQ]:=Module[{O},O=Ket[a[[1]]];
L=Length[a];
Do [O=O\[CircleTimes]Ket[a[[i]]],{i,2,L}];O];

BraV[a_?ListQ]:=Module[{O},O=Bra[a[[1]]];
L=Length[a];
Do [O=O\[CircleTimes]Bra[a[[i]]],{i,2,L}];O];

(* Generate decimal Xd and binary Xb arrays for nq qubits *)
(* Xd:=Array[xd,nb^nq ,0]; Xb=Array[xb,nb^nq ,0]; *)

xd[i_]:=i;
xb[i_]:=IntegerDigits[xd[i],nb,nq];
Xd[n_]:=Array[xd,n,0]; 
Xb[n_]:=Array[xb,n,0];

(*  Define: Adjoint as Adj *)
 
Adj[x_?MatrixQ] := Transpose[Conjugate[x]]; 

(*  Define: Four Pauli Matrices in three ways  *) 

s[0]:={{1,0},{0,1}};
s[1]:={{0,1},{1,0}};
s[2]:={{0,-I},{I,0}};
s[3]:={{1,0},{0,-1}};

Sigma0:={{1,0},{0,1}};
Sigma1:={{0,1},{1,0}};
Sigma2:={{0,-I},{I,0}};
Sigma3:={{1,0},{0,-1}};

(* The Greek letter version *)
\!\(\[Sigma]\_0\  := \ Sigma0\);
\!\(\[Sigma]\_1\  := \ Sigma1\);
\!\(\[Sigma]\_2\  := \ Sigma2\);
\!\(\[Sigma]\_3\  := \ Sigma3\);

(* The Doublestruck I is the 2x2 Unit Matrix *)

\[DoubleStruckCapitalI] := Sigma0;
 
(*  Define:  Tensor Product of two matrices *)

  x_?MatrixQ \[CircleTimes] y_?MatrixQ :=BlockMatrix[Outer[Times,x,y]];
 
(*  TRACES *)

(* Partial Trace over qubit q of Spin Products  *)
(* ZeroM[s_]:=Table[0,{i,s},{j,s}]//MatrixForm; *)

PTrS[q_, L_,V_]:=If[V[[q]]==0,2* SP[L-1,Drop[V,{q}]],ZeroMatrix[2^(L-1)]];

(*  Partial Trace Over general 2^L X 2^L operator Omega *)

PTr[TO_ /; VectorQ[TO], \[CapitalOmega]_ /; MatrixQ[\[CapitalOmega]]] := 
      Module[{nto, n\[CapitalOmega], W, VOP, WP, V, V2}, nto = \
Length[TO];
      n\[CapitalOmega] = Log[2, Length[\[CapitalOmega]]];
      If[nto == n\[CapitalOmega], X = \
Tr[\[CapitalOmega]], W = Array[w, n\[CapitalOmega]];
        VOP = DeleteCases[ReplacePart[Array[w, n\[CapitalOmega]], \
z, Split[TO]], z];
        WP = ReplacePart[W, 0, Split[TO]];
        V = Array[{{w[#], 0, 3}} &, n\[CapitalOmega]];
        Do[{V = MapAt[Apply[Sequence, #] &, V, {i}]}, {i, 1, n\[CapitalOmega] - nto}];
        V = FlattenAt[V, n\[CapitalOmega]];
        V2 = DeleteCases[ReplacePart[V, z, Split[TO]], z];
        V2 = Partition[Flatten[V2], 3];
        X = Sum[2^(nto - n\[CapitalOmega])*Tr[\[CapitalOmega].SP[n\[CapitalOmega], WP]]*SP[n\[CapitalOmega] - nto, VOP], 
        Evaluate[Sequence @@ V2]]]; Simplify[X]];

(* compiled version of PTr *)
CompPTr= Compile[{{TO,_Integer,1},{\[CapitalOmega],_Complex,2}}, 
PTr[TO, \[CapitalOmega] ], {{PTr[TO, \[CapitalOmega] ] , _Complex, 2}}]  ;


(*  Partial Trace Over general 2^L X 2^L operator Omega  a second, sometimes faster, version *)
(*  L is  the number of qubits; outList is an array which stipulates the \
traced-out qubits;\[CapitalOmega] is the matrix uppon which the \
partial trace is perfomed *) 

\!\(\(PTrGen[TO_ /; VectorQ[TO], \[CapitalOmega]_ /; MatrixQ[\
\[CapitalOmega]]] :=
 Module[{nto,n\[CapitalOmega]}, nto = Length[TO];
 n\[CapitalOmega] = Log[2,Length[\[CapitalOmega]]]; \
numLeft = n\[CapitalOmega] - nto;
 If[numLeft > 0, {plugZeros[zeroPositions_, W_] :=
 If[Length[zeroPositions] > 1,plugZeros[Drop[zeroPositions, 1], 
  Insert[W, 0, zeroPositions[\([1]\)]]], Insert[W, \
0,zeroPositions[\([1]\)]]]; 
coef[outList1_, \[CapitalOmega]1_?MatrixQ, W_]  := 
Tr[\[CapitalOmega]1 . SP[Log[2, \
Length[\[CapitalOmega]1]], plugZeros[Sort[outList1], W]]]; 
Base4\_{j_,l_}[k_]:= \(Base4\_{j,l}[k]=Floor[\((j-4^\((l-k+1)\)*Floor[j/\((4^\
\((l-k+1)\))\)])\)/4^\((l-k)\)]\);
X=Sum[coef[TO,\[CapitalOmega],Array[Base4\_{i,numLeft},numLeft]]*\
SP[numLeft,Array[Base4\_{i,numLeft}, 
 numLeft]], {i,0, 4^\((numLeft)\) - 1}]\ /2^numLeft},X = Tr[\[CapitalOmega]]];
 Simplify[X]];\)\)

(* Multi qubit measurement on the computational bases on a vector *)
(* Input:
        O: original vector with length 2^NQ
       MQ: List of measured qubits, e.g., {1,3,6,7,43}
       MV: Measured values, e.g., {0,0,1,1,0}  
   Output: OutV/Norm[OutV], normalized vector of NQ-MQ qubits  *)

MeasureKet[O_?ListQ,MQ_?ListQ,MV_?ListQ]:= Module[{},
ON=Length[O];     (*Length of vector*)
LM=Length[MQ];    (*Number of measured Qubits*)
LMV=Length[MV];    (*Number of measured Qubits*)
NQ=Log[2,ON];     (*Total Number of qubits*)
ListN=Table[i,{i,1,NQ}];  (*List of all qubits *)
OutV=Table[0,{i,1,2^(NQ-LM)}];  (* Initialized output vector*)
MM=Complement[ListN,MQ];        (* List of not measured qubits*)
Seq=Array[{{w[MM[[#]]],0,1}}&,Length[MM]];  (* Array to sum over non measured qubits*)
Seq=Partition[Flatten[Seq],3];
MesNumber=Sum[MV[[i]]*2^(NQ-MQ[[i]]),{i,1,LM}]; (* construction of measured number*)
NonMeas=Sum[w[MM[[i]]]*2^(NQ-MM[[i]]),{i,1,NQ-LM}]; (* Non measured contribution *)
Posit=Sum[w[MM[[i]]]*2^(NQ-Length[MQ]-i),{i,1,NQ-LM}]+1;  (* Position in the output vector *)
Do [OutV=ReplacePart[OutV,O[[MesNumber+NonMeas+1]],Posit],Evaluate[Sequence@@Seq]];
If[Norm[OutV]\[NotEqual]0,OutW=OutV/Norm[OutV],OutW=ZeroT[Length[OutV]]];
OutW
]

CompMeasureKet=Compile[{{A,_Complex,1},{M,_Integer,1},{vM,_Integer,1}},
MeasureKet[A,M,vM ], {{MeasureKet[A,M,vM], _Complex, 1}}]  ;

(*  ================================================================= *)
(* UNITARY OPERATORS  *)
(* Define Time dependent Hadamard  *)
(* hadT[n_]:=  *)
 
 
(* -- Define:  Tensor Products for 2  qubits *)
TensorProduct[x_?MatrixQ,y_?MatrixQ]:=BlockMatrix[Outer[Times,x,y]];

(*-- Define:  Commutator for Matrices x and y  *)
Comm[x_?MatrixQ,y_?MatrixQ]:=  x.y - y.x
     
(* We count qubits with convention leftmost binary=qubit 1, rightmost \
binary=qubit nq *)
(* i,j are the decimal numbers for the state of all nq qubits *)
(* Faster Recursion method--with array labels *)

SP[L_, Q_] := SP[L, Q] = If [L > 1, TensorProduct[SP[L - 1,Drop[Q,-1]],s[Q[[L]]] ],s[Q[[1]]]];

(* *)
(* GATES *)
(*  Build operators acting on two qubits(q1,q2) TwoOP, 
and three qubits(q1,q2,q3 ) ThreeOP *)
(* TwoOp and ThreeOp are used to build general 
2-qubit and 3-qubit Gates *)
(* ZeroT  is an array of L zeroes corresponding 
to all L operators as unitoperators *)

ZeroT[L_]:=Table[0,{L}];

OneOp[L_, q1_, op1_] := SP[L, ReplacePart[ZeroT[L], op1, q1]];

TwoOp[L_,q1_,q2_,op1_,op2_]:=SP[L,ReplacePart[ReplacePart[ZeroT[L],op1,q1],op2,q2]];

ThreeOp[L_,q1_,q2_,q3_,op1_,op2_,op3_]:=
SP[L,ReplacePart[ReplacePart[ReplacePart[ZeroT[L],op1,q1],op2,q2],op3,q3]];

(* Controlled-X ,Controlled-Y, and Controlled-Z gates. Involves 2 qubits  *)
 
ControlledX[L_,q1_,q2_]:=(TwoOp[L,q1,q2,0,0]+TwoOp[L,q1,q2,3,0]+TwoOp[L,q1,q2,0,1]-TwoOp[L,q1,q2,3,1])/2;
 
ControlledY[L_,q1_,q2_]:=(TwoOp[L,q1,q2,0,0]+TwoOp[L,q1,q2,3,0]+TwoOp[L,q1,q2,0,2]-TwoOp[L,q1,q2,3,2])/2;
 
ControlledZ[L_,q1_,q2_]:=(TwoOp[L,q1,q2,0,0]+TwoOp[L,q1,q2,3,0]+TwoOp[L,q1,q2,0,3]-TwoOp[L,q1,q2,3,3])/2;

(* CNOT gate (same as ControlledX), Involves 2 qubits, ith qubit is control, \
jth qubit is target *)

CNOT[L_,q1_,q2_]:=(TwoOp[L,q1,q2,0,0]+TwoOp[L,q1,q2,3,0]+TwoOp[L,q1,q2,0,1]-TwoOp[L,q1,q2,3,1])/2 ;

(* CPHASE GATE *)
(* Matrix case, to act on density matrices *)
CPHASE[L_, q1_, q2_] := (TwoOp[L, q1, q2, 0, 0]
 + TwoOp[L, q1, q2, 3, 0] +TwoOp[L, q1, q2, 0, 3]
 - TwoOp[L, q1, q2, 3, 3])/2;

(* Vector case, to act on a general vector with 2^nqL components *)

CPHASEV[i_, j_, vec_?ListQ] := Module[{NN, vat}, 
NN = Length[vec]; 
nqL = Log[2,NN];
vat = vec; 
Do [If[IntegerDigits[num,2,nqL][[i]] == 1 &&IntegerDigits[num,2,nqL][[j]] == 1, vat[[num + 1]] = -vec[[num + 1]] ], {num, 0,NN - 1}]; vat];

(* Swap  gate. Involves 2 qubits q1 q2 out of the L qubits  *)

Swap[L_,q1_,q2_]:=CNOT[L,q1,q2].CNOT[L,q2,q1].CNOT[L,q1,q2];
 
(* Define scriptH as Hadamard  acting on one qubit  Can use \[CircleTimes] \
forms to build many qubit operators along with
the Doublestruck I --the 2x2 Unit Matrix  *)

\[ScriptCapitalH] = (s[1] + s[3])/Sqrt[2]  ;

(* Define Hadamard acting on one qubit q out of L qubits *)

had[L_,q_]:= If[q>0,(SP[L, ReplacePart[ZeroT[L],3,q]] + \
SP[L,ReplacePart[ZeroT[L],1, q]])/Sqrt[2],s[0] ];

(* Define Hadamard  acting on a qubit set Q={q1,q2,...} out of L qubits *)

(* Hadamard  "Had[L,Q]"   acts on  all members of set Q={q1,q2,...} of the L qubits *)

Had[L_, Q_] := Had[L, Q] =If [L > 1, TensorProduct[Had[L - 1,Drop[Q,-1]],had[1,Q[[L]]] ], had[1,Q[[1]]]];

(* Hadamard  HALL has a hadamard acting on all members of set Q={q1,q2,...} of the L qubits *)

HALL[L_] := HALL[L] =Module[{Q},Q=Table[1,{L}];
If [L > 1, TensorProduct[Had[L - 1,Drop[Q,-1]],had[1,Q[[L]]] ], had[1,Q[[1]]]]];

(* Toffoli gate. Involves 3 qubits,ith qubit is control, jth qubit is target *)

Toffoli[L_,q1_,q2_,q3_]:=
(3*ThreeOp[L,q1,q2,q3,0,0,0]+ThreeOp[L,q1,q2,q3,3,0,0]+ThreeOp[L,q1,q2,q3,0,3,0]
-ThreeOp[L,q1,q2,q3,3,3,0]+ThreeOp[L,q1,q2,q3,0,0,1]-ThreeOp[L,q1,q2,q3,3,0,1]
-ThreeOp[L,q1,q2,q3,0,3,1]+ThreeOp[L,q1,q2,q3,3,3,1])/4;

(* Define Projection operators (for a=0 |0><0|; a=1 |1><1|)  acting on one qubit q out of L qubits *) 
(*  proj[L,q,a]  gives the projection operator P_a  acting on qubit q out of \
the set of L qubits; it is the L-qubit operator 1X1XP_0 X 1  for q=3,a=0 and \
L=4;  Here 1=s[0] unit 2x2 matrix *)
  
proj[L_,q_,a_]:= (SP[L, ReplacePart[ZeroT[L],0,q]] + (-1)^a SP[L,ReplacePart[ZeroT[L],3, q]])/2;


(* Projection Operators "PJ[L,Q,A]"  which acts on set {q1,q2,...} of the L \
qubits *)
(* A denotes an array{a1....aL}  indicating the projection to either zero or \
one; if ai is not equal to zero or one then the operator 1=s[0] is taken for \
the ith qubit  *)
(* Thus one can build an operator of the type:  1 X P_0 X P_1 X 1 X P_0 X P_1 \
X 1,  which asks if the qubits 2,3,5,6 have the values 0101  *)

PJ[1, {a_}] := (s[0] + (-1)^a  s[3])/2;
PJ[L_, A_] := PJ[L, A] = TensorProduct[PJ[L - 1, Drop[A, -1]], PJ[1, Drop[A, \
L - 1]]];
(* Define Projection operators as (\[ScriptCapitalP]\_a_\)  to be used in  \
\[CircleTimes] forms *) 

\!\(\(\(\[ScriptCapitalP]\_a_\ \  := PJ[1, {a}];\)\(\n\) \)\)

PX[0]  := TensorProduct[KetX[0],BraX[0]];
PX[1]  := TensorProduct[KetX[1],BraX[1]];

PY[0]  := TensorProduct[KetY[0],BraY[0]];
PY[1]  := TensorProduct[KetY[1],BraY[1]];
(* General Projection operators *)

ProjG[a_,v_?ListQ]:=Module[{w},
w=v/Norm[v];
 (s[0]+(-1)^a*(w[[1]]*s[1]+w[[2]]*s[2]+w[[3]]*s[3]))/2];
 
MultiProjCBasis[a_?ListQ]:=Module[{\[CapitalOmega]},
L=Length[a];
\[CapitalOmega]=PJ[1,{a[[1]]}];
Do [\[CapitalOmega]=\[CapitalOmega]\[CircleTimes] PJ[1,{a[[i]]}]  ,{i,2,L}];\[CapitalOmega]];

MultiQProj[aa_?ListQ,bb_?ListQ]:=Module[{}, 
la=Length[bb];
cc:=Table[ReplaceAll[aa[[i]],{0\[Rule]s[0],1\[Rule]PX[bb[[i]]],2\[Rule]PY[bb[[i]]],3\[Rule]PJ[1,{bb[[i]]}]}],{i,1,la}];
resul[1]:=cc[[1]];
resul[j_]:=(resul[j-1]\[CircleTimes]cc[[j]]);
resul[la]
];

(*  Generate Ket *)

Ket1[a_,b_]:=Cos[a/2]*Exp[-\[ImaginaryI]*b/2]*Ket[0]+Sin[a/2]*Ket[1]*Exp[-\[ImaginaryI]*b/2]   ;

Bra1[a_,b_]:=Cos[a/2]*Exp[-\[ImaginaryI]*b/2]*Bra[0]+Sin[a/2]*Bra[1]*Exp[-\[ImaginaryI]*b/2]   ;


RandomKet1 := 
    Module[{\[Theta], \[Phi]}, 
       \[Theta] = Random[Real, {0, Pi}]; 
       \[Phi]   = Random[Real, {0, 2*Pi}]; 
      OKet = Cos[\[Theta]/2]*Exp[-\[ImaginaryI]*\[Phi]/2]*Ket[0]+Sin[\[Theta]/2]*Ket[1]*Exp[-\[ImaginaryI]*\[Phi]/2]     ]

(*  Generate a Bra *)

RandomBra1 := 
    Module[{\[Theta], \[Phi]}, 
       \[Theta] = Random[Real, {0, Pi}]; 
       \[Phi]   = Random[Real, {0, 2*Pi}]; 
      OKet = Cos[\[Theta]/2]*Exp[-\[ImaginaryI]*\[Phi]/2]*Bra[0]+Sin[\[Theta]/2]*Bra[1]*Exp[-\[ImaginaryI]*\[Phi]/2]     ]
(*  Generate a random 1 qubit state *)

RandomQubit1 := 
    Module[{\[Theta], \[CurlyPhi], P,Px,Py,Pz}, 
       \[Theta] = Random[Real, {0, Pi}]; 
       \[CurlyPhi] = Random[Real, {0, 2*Pi}]; 
      P = Random[Real, {0, 1}]; Px = P*Sin[\[Theta]]*Cos[\[CurlyPhi]]; 
      Py = P*Sin[\[Theta]]*Sin[\[CurlyPhi]]; 
      Pz = P*Cos[\[Theta]]; 
      \[Rho] = (1/2)*( s[0] + Px*s[1] + Py*s[2] + Pz*s[3] )]

(*  Generate a random 2 qubit state *)

RandomQubit2:= Module[{i,j,U},
RandomQubit1;A=\[Rho];
RandomQubit1;B=\[Rho];
Do[J[i,j]=Random[Real,{-100,100}],{i,1,3},{j,1,3}];
U=MatrixExp[Sum[\[ImaginaryI]*J[i,j]*(s[i]\[CircleTimes] s[j]),{i,1,3},{j,1,3}]];
\[Rho]12=Chop[Adj[U].(A\[CircleTimes]B).U];\[Rho]12]

(*  Generate a random N qubit state *)

RandomQubitN[L_]:=Module[{AAA,BB,MM,Seq,indc,w},
Do[AAA[i] = RandomQubit1, {i, 1, L}];
BB[1] := AAA[1];
BB[i_] := BB[i]=(AAA[i]\[CircleTimes]BB[i - 1]);
MM = Table[i, {i, 1, L}];
Seq = Array[{{w[MM[[#]]], 0, 1}} &, L];
Seq = Partition[Flatten[Seq], 3];
indc = Table[w[MM[[i]]], {i, 1, L}];
Do[J[Flatten[indc]] = Random[Real, {-100, 100}], Evaluate[Sequence @@ Seq]];
U = MatrixExp[
      Sum[\[ImaginaryI]*J[Flatten[indc]]*SP[L, indc], Evaluate[Sequence @@ Seq]]];
\[Rho]N = Chop[Adj[U].(BB[L]).U];
Chop[\[Rho]N]];

(*  Oracle for x=nqx qubits and y= nqy qubit case *)
(*  Uf | x y> = |x y+f(x)>  *)

(* Definef defines the function:itype=1 is the random function case,itype=2  \
is a Grover selection case *)

(* Definef defines the function : itype = -1 is the random function 
case, itype = -2 is a Grover selection case;   *)
 

Definef[nx_, ny_, itype_] :=
 Module[{xv}, Nv = itype; xv[i_] := Mod[Xd[nx*ny][[i]], nx];
 If[itype == -1, f = Table[Random[Integer, {0, ny - 1}], {i, nx}], 
 If[itype == -2, f =ReplacePart[ZeroT[nx], 1, xpick = Random[Integer, \
{1,nx}]], 
f = Table[Mod[xguess^xv[i], Nv], {i, nx}]]]];

Oracle[nqx_, nqy_, itype_] := Module[{xv, yv, nx, ny, n, i},nx = nb^nqx; ny = \
nb^nqy; n = nx*ny;
Definef[nx, ny, itype];
xv[i_] := FromDigits[ Take[IntegerDigits[i, 2, nqx + nqy] , nqx], 2];
 yv[i_] := FromDigits[Take[IntegerDigits[i, 2, nqx + nqy], {nqx + 1,nqx + \
nqy}], 2];
Table[KroneckerDelta[xv[i],xv[j]]*KroneckerDelta[yv[i],Mod[yv[j]+f[[xv[i]+1]],\
ny]],{i,0,n-1},{j,0,n-1}]];

(*Purity*) 
Purity[A_?MatrixQ]:=Chop[Tr[A.A]];



(* Rotation around a vector v an angle \[Theta] *)

Rotqbit[v_?ListQ,t_]:=MatrixExp[-\[ImaginaryI]*t/2*(s[1]*v[[1]]+s[2]*v[[2]]+s[3]*v[[3]])];

(* Rotation around X, Y, Z an angle t *)

RotX[t_]:=MatrixExp[-\[ImaginaryI]*t/2*s[1]];
RotY[t_]:=MatrixExp[-\[ImaginaryI]*t/2*s[2]];
RotZ[t_]:=MatrixExp[-\[ImaginaryI]*t/2*s[3]];

(* Entropy *)

Entropy[O_?MatrixQ]:= Module[{},
M = Chop[Eigenvalues[O]];
MOPO = DeleteCases[M, 0 ];
OLOGO = -Sum[MOPO[[i]]*Log[2, MOPO[[i]]], {i, 1, Length[MOPO]}];
OLOGO]


(* Fidelity *)

Fidelity[O_?MatrixQ,P_?MatrixQ]:= Module[{},
SQO=MatrixPower[O,1/2];
FideM=MatrixPower[SQO . P . SQO,1/2];
Chop[Tr[FideM]]
]

(*  =================================================================  *)

End[]

EndPackage[];


